/*
    DB Port
    Copyright (C) 2010 Daniel Mewes
*/

package dbport;

import java.util.*;
import java.text.*;

import systeminterface.*;
import metadata.*;
import operator.*;
import relationalalgebra.*;
import util.*;
import exceptions.*;

class QueryDataPage extends HtmlPage {
    private static PredicateTreeNode getPredicateFromArguments(Map<String, String> argumentMap, Map<String, Type> schema, String nodeId) throws Exception {
        String predicateNodeType = argumentMap.get("predicate" + nodeId);
        String predicateNodeLeft = argumentMap.get("predicate" + nodeId + "Left");
        String predicateNodeRight = argumentMap.get("predicate" + nodeId + "Right");
        
        if (predicateNodeType == null || predicateNodeType.equals("null"))
            return null;
        else {
            if (predicateNodeType.equals("OR")) {
                return new PrimitivePredicateTreeNode(getPredicateFromArguments(argumentMap, schema, predicateNodeLeft), LogicalOperator.OR, getPredicateFromArguments(argumentMap, schema, predicateNodeRight));
            }
            else if (predicateNodeType.equals("AND")) {
                return new PrimitivePredicateTreeNode(getPredicateFromArguments(argumentMap, schema, predicateNodeLeft), LogicalOperator.AND, getPredicateFromArguments(argumentMap, schema, predicateNodeRight));
            }
            else {
                String value = predicateNodeRight;
                Object valueObject = null;
                
                Type valueType = schema.get(predicateNodeLeft);
                if (valueType.getName().equals("Integer"))
                    valueObject = new Integer(value);
                else if (valueType.getName().equals("Long"))
                    valueObject = new Long(value);
                else if (valueType.getName().equals("Float"))
                    valueObject = new Float(value);
                else if (valueType.getName().equals("Double"))
                    valueObject = new Double(value);
                else if (valueType.getName().equals("Date"))
                    valueObject = (new SimpleDateFormat("yyyy-MM-DD")).parse(value);
                else // Hope we have a String type...
                    valueObject = value;
                
                if (predicateNodeType.equals("EQ"))
                    return new PrimitivePredicateTreeNode(predicateNodeLeft, ComparisonOperator.EQ, valueObject);
                else if (predicateNodeType.equals("NEQ"))
                    return new PrimitivePredicateTreeNode(predicateNodeLeft, ComparisonOperator.NEQ, valueObject);
                else if (predicateNodeType.equals("LT"))
                    return new PrimitivePredicateTreeNode(predicateNodeLeft, ComparisonOperator.LT, valueObject);
                else if (predicateNodeType.equals("LEQ"))
                    return new PrimitivePredicateTreeNode(predicateNodeLeft, ComparisonOperator.LEQ, valueObject);
                else if (predicateNodeType.equals("GT"))
                    return new PrimitivePredicateTreeNode(predicateNodeLeft, ComparisonOperator.GT, valueObject);
                else if (predicateNodeType.equals("GEQ"))
                    return new PrimitivePredicateTreeNode(predicateNodeLeft, ComparisonOperator.GEQ, valueObject);
                else
                    return null;
            }            
        }
    }
    
    private static RelationalAlgebraExpression getQueryFromArguments(Map<String, String> argumentMap, String nodeId) throws Exception {
        if (nodeId == null) {
            return getQueryFromArguments(argumentMap, argumentMap.get("resultNode"));
        }
        else {
            int nodeNumber = Integer.decode(nodeId).intValue();
            
            if (nodeNumber >= 1 && nodeNumber <= 7) // Input
                return new Input(argumentMap.get("input" + nodeNumber + "Table"));
            else if (nodeNumber >= 8 && nodeNumber <= 13) // Join
                return new Join(getQueryFromArguments(argumentMap, argumentMap.get("join" + nodeNumber + "Node1")), getQueryFromArguments(argumentMap, argumentMap.get("join" + nodeNumber + "Node2")), argumentMap.get("join" + nodeNumber + "Attribute1"), argumentMap.get("join" + nodeNumber + "Attribute2"));
            else if (nodeNumber >= 14 && nodeNumber <= 19) // Cross product
                return new CrossProduct(getQueryFromArguments(argumentMap, argumentMap.get("crossProduct" + nodeNumber + "Node1")), getQueryFromArguments(argumentMap, argumentMap.get("crossProduct" + nodeNumber + "Node2")));
        }
    
        return null;
    }
    
    private static void extractSchema(RelationalAlgebraExpression node, StorageLayer storageLayer, Map<String, Type> schema) throws Exception {
        switch (node.getType()) {
            case INPUT:
                Table table = storageLayer.getTableByName(((Input)node).getRelationName());
                Operator<? extends Column> columnOperator = table.getAllColumns();
                columnOperator.open();
                Column currentColumn;
                while ((currentColumn = columnOperator.next()) != null) {
                    schema.put(currentColumn.getColumnName(), currentColumn.getColumnType());
                }
                columnOperator.close();
                break;
            case JOIN:
                extractSchema(((Join)node).getLeftInput(), storageLayer, schema);
                extractSchema(((Join)node).getRightInput(), storageLayer, schema);
                break;
            case SELECTION:
                extractSchema(((Selection)node).getInput(), storageLayer, schema);
                break;
            case PROJECTION:
                extractSchema(((Projection)node).getInput(), storageLayer, schema);
                break;
            case CROSS_PRODUCT:
                extractSchema(((CrossProduct)node).getLeftInput(), storageLayer, schema);
                extractSchema(((CrossProduct)node).getRightInput(), storageLayer, schema);
                break;
            default:
                System.err.println("Unhandled relation algebra node");
        }
    }

    public static String getContent(DbEnvironment environment, String[] arguments) throws Exception {
        StringBuffer result = new StringBuffer();
        
        result.append(getHtmlHead());
        
        if (arguments.length > 0) {
            // Try to query the database
            
            String repeat = null;
            String projection = null;
            String delete = null;
            
            Map<String, String> argumentMap = new HashMap<String, String>();
            
            for (String arg : arguments) {
                if (arg.length() >= 7 && arg.substring(0, 7).equals("repeat="))
                    repeat = arg.substring(7);
                else if (arg.length() >= 11 && arg.substring(0, 11).equals("projection="))
                    projection = arg.substring(11);
                else if (arg.length() >= 7 && arg.substring(0, 7).equals("delete="))
                    delete = arg.substring(7);
                else {
                    String[] splittedArgument = arg.split("=", 2);
                    argumentMap.put(splittedArgument[0], splittedArgument[1]);
                }
            }
            
            // Build query
            RelationalAlgebraExpression query = getQueryFromArguments(argumentMap, null);
            
            // Build schema
            Map<String, Type> schema = new HashMap<String, Type>();
            extractSchema(query, environment.database.getStorageInterface(), schema);
            
            // Build predicate
            PredicateTreeNode predicate = getPredicateFromArguments(argumentMap, schema, "root");
            
            // Add Selection
            if (predicate != null)
                query = new Selection(query, predicate);
            
            // Add projection
            if (projection.length() > 0) {
                String[] projectionAttributes = projection.split(",");
                for (int i = 0; i < projectionAttributes.length; ++i)
                    projectionAttributes[i] = projectionAttributes[i].trim();
                query = new Projection(query, projectionAttributes);
            }
            
            // This check actually comes a bit too late...
            if (query != null && projection != null) {   
                int repetitions = 1;
                if (repeat != null) {
                    repetitions = Integer.decode(repeat).intValue();
                }
                                         
                result.append("<h1>Query database</h1>\n");

                result.append("<h2>Query plan</h2>");
                // Ugly hack to preserver formatting...
                result.append("<textarea rows=\"10\" cols=\"140\">");
                result.append(environment.database.getQueryInterface().explain(query).replace("<", "&lt;").replace(">", "&gt;"));
                result.append("</textarea>");
                
                result.append("<h2>Query result</h2>");
                
                long startTime = System.currentTimeMillis();
                Operator<? extends Row> queryResult = null;
                while (repetitions > 0) {
                    queryResult = environment.database.getQueryInterface().query(query);
                
                    repetitions--;
                    
                    queryResult.open();
                    if (delete != null && delete.equals("true")) { // Delete rows
                        String tableName = null;
                        // Only works for single-input / no-projection queries
                        if (query instanceof Selection)
                            tableName = ((Input)((Selection)query).getInput()).getRelationName();
                        else
                            tableName = ((Input)query).getRelationName();
                            
                        Row currentRow = null;
                        while ((currentRow = queryResult.next()) != null) {
                            try {
                                environment.database.getQueryInterface().deleteRow(tableName, currentRow);
                            } catch (Exception e) {
                                e.printStackTrace();
                            }
                            // We may have to requery to not destroy the operator (implementation dependant)
                            queryResult.close();
                            queryResult = environment.database.getQueryInterface().query(query);
                            queryResult.open();
                        }
                        
                        result.append("<h2>Rows deleted</h2>");
                    }
                    else if (repetitions == 0) { // Output result
                        int rowCount = 0;
                        Row currentRow = null;
                        boolean firstRow = true;
                        String[] columns = null;
                        result.append("<table border=\"1\">\n");
                        while ((currentRow = queryResult.next()) != null) {
                            rowCount++;
                            result.append("<tr>");
                            if (firstRow) {
                                firstRow = false;
                                columns = currentRow.getColumnNames();
                                for (String column : columns)
                                    result.append("<th>" + column + "</th>");
                                result.append("</tr>\n<tr>");
                            }
                            for (String column : columns) {
                                Object value = currentRow.getColumnValue(column);
                                if (value == null)
                                    result.append("<td>&lt;null&gt;</td>");
                                else
                                    result.append("<td>" + value.toString() + "</td>");
                            }
                            result.append("</tr>\n");
                        }
                        result.append("</table><br>\nGot " + rowCount + " rows<br>\n");
                    }
                    else { // Just iterate over the result, to force evaluation
                        Row currentRow = null;
                        boolean firstRow = true;
                        String[] columns = null;
                        while ((currentRow = queryResult.next()) != null) {
                            if (firstRow) {
                                firstRow = false;
                                columns = currentRow.getColumnNames();
                            }
                            for (String column : columns)
                                currentRow.getColumnValue(column);
                        }
                    }
                    queryResult.close();
                }
                
                result.append("<br>Took " + (System.currentTimeMillis() - startTime) + " ms<br>");
                
            }
            else
                result.append("<h1>Missing argument</h1>");
        }
        else {
            result.append("<h1>Query database</h1>\n");
            
            result.append("<form method=\"get\" action=\"#\">\n");
            
            result.append("Run query <input name=\"repeat\" size=\"8\" value=\"1\"> times<br><br>\n");
            
            result.append("<table border=\"0\">\n");
            
            result.append("<tr><td>Input nodes: </td><td>");
            // Only 4 are officially supported, but more may be interesting for testing purposes...
            for (int i = 1; i <= 7; ++i) {
                result.append("Input " + i + ": ");
                result.append("<input name=\"input" + i + "Table\" size=\"16\" value=\"table name\"><br>");
            }
            result.append("</td></tr>\n");
            
            result.append("<tr><td>Join nodes: </td><td>");
            // Only 3 are officially supported, but more may be interesting for testing purposes...
            for (int i = 8; i <= 13; ++i) {
                result.append("Join " + i + ": ");
                result.append("JOIN <input name=\"join" + i + "Node1\" size=\"16\" value=\"left node number\"> ");
                result.append("WITH <input name=\"join" + i + "Node2\" size=\"16\" value=\"right node number\"> ");
                result.append("ON <input name=\"join" + i + "Attribute1\" size=\"16\" value=\"left attribute\">");
                result.append("=<input name=\"join" + i + "Attribute2\" size=\"16\" value=\"right attribute\"><br>");
            }
            result.append("</td></tr>\n");
            
            result.append("<tr><td>Cross product nodes: </td><td>");
            // Only 3 are officially supported, but more may be interesting for testing purposes...
            for (int i = 14; i <= 19; ++i) {
                result.append("Cross product " + i + ": ");
                result.append("<input name=\"crossProduct" + i + "Node1\" size=\"16\" value=\"node number\"> ");
                result.append("X <input name=\"crossProduct" + i + "Node2\" size=\"16\" value=\"node number\"><br>");
            }
            result.append("</td></tr>\n");
            
            result.append("<tr><td>Take result from node...</td><td>");
            result.append("<input name=\"resultNode\" size=\"16\" value=\"node number\"> ");
            result.append("</td></tr>\n");
            
            result.append("<tr><td>...apply predicate...</td><td>");
            result.append("Predicate root: ");
            result.append("<input name=\"predicaterootLeft\" size=\"16\" value=\"node number / column\"> ");
            result.append("<select name=\"predicateroot\"><option value=\"null\">-</option><option value=\"AND\">AND</option><option value=\"OR\">OR</option><option value=\"EQ\">=</option><option value=\"NEQ\">!=</option><option value=\"LT\">&lt;</option><option value=\"LEQ\">&lt;=</option><option value=\"GT\">&gt;</option><option value=\"GEQ\">&gt;=</option></select>");
            result.append("<input name=\"predicaterootRight\" size=\"16\" value=\"node number / value\"><br>");
            for (int i = 1; i <= 20; ++i) {
                result.append("Predicate node " + i + ": ");
                result.append("<input name=\"predicate" + i + "Left\" size=\"16\" value=\"node number / column\"> ");
                result.append("<select name=\"predicate" + i + "\"><option value=\"null\">-</option><option value=\"AND\">AND</option><option value=\"OR\">OR</option><option value=\"EQ\">=</option><option value=\"NEQ\">!=</option><option value=\"LT\">&lt;</option><option value=\"LEQ\">&lt;=</option><option value=\"GT\">&gt;</option><option value=\"GEQ\">&gt;=</option></select>");
                result.append("<input name=\"predicate" + i + "Right\" size=\"16\" value=\"node number / value\"><br>");
            }
            result.append("</td></tr>\n");
            
            result.append("<tr><td>...and project to: </td><td>");
            result.append("<input name=\"projection\" size=\"32\" value=\"column1, column2, ...\"> ");
            result.append("</td></tr>");
                   
            result.append("</table>\n");
            
            result.append("<br><br><input type=\"checkbox\" name=\"delete\" value=\"true\"> Delete selected rows (only supported for single-input queries without projection; slow implementation! do not use for benchmarking)<br>");
            
            result.append("<input type=\"submit\">\n");
            result.append("</form>\n");
        }
        
        result.append(getHtmlTail());
        
        return result.toString();
    }
}
